# $Id: server.py 3020 2010-02-08 11:51:58Z wsl $
# $URL: https://svn.non-gnu.uvt.nl/uvt-dev/tags/sources/meresco/bzv-8.2.0/src/server.py $
import pprint
import gc

infra = None

from merescocore.components.statistics import Statistics, Logger

from merescocore.components import Lxml2Amara, XmlXPath, XmlPrintLxml, Amara2Lxml, Venturi, Configuration, FilterField, RenameField, TransformFieldValue, RewritePartname, Xml2Fields, XPath2Field, FilterMessages, RssItem, Fields2XmlTx, Reindex, ReindexConsole, XsltCrosswalk
#from merescocore.components import StorageComponent, Lxml2Amara, XmlXPath, XmlPrintLxml, Amara2Lxml, Venturi, Configuration, FilterField, RenameField, TransformFieldValue, RewritePartname, Xml2Fields, XPath2Field, FilterMessages, RssItem, Fields2XmlTx, Reindex, ReindexConsole, XsltCrosswalk
from merescocore.components.statisticsxml import StatisticsXml
from merescocore.components.drilldown import SRUTermDrilldown, DrilldownFieldnames
from merescocore.components.http import PathFilter, FileServer, PathRename
#from merescocore.components.http.webrequestserver import WebRequestServer
from merescocore.components.sru import SruParser, SruHandler, SRURecordUpdate
from merescocore.components.tokenizefieldlet import TokenizeFieldlet

from merescocore.framework import Transparant, TransactionScope,  BatchTransactionScope, ResourceManager, Observable

from merescocomponents.facetindex import Drilldown, LuceneIndex, CQL2LuceneQuery, Fields2LuceneDocumentTx
from merescocomponents.facetindex.merescolucene import IndexWriter, merescoStandardAnalyzer
from merescocomponents.facetindex.tools import unlock

# ImportError: No module named ngram
#from merescocomponents.ngram import NGramFieldlet, NGramQuery, RatioSuggester

from os.path import join, dirname, abspath, isdir
from os import remove, makedirs, unlink
from glob import iglob

## BZV specifics ###

# Anton: use homebrew dynamichtml from bzv instead of the cq2 version to support logging
from bzv.dynamichtml import DynamicHtml

# Anton: Note: 2 versions of delfttilburg!
from bzv.delfttilburg.normalize import LastnameNormalize, FirstInitialNormalize
from delfttilburg import RateRecordCookie, RateRecord, MultiLevelDrilldown, ComposeNameFromMods, ModsDateField, DidYouMean
from bzv.drilldownconfig import DrilldownConfig

from bzv.yearnormalize import YearNormalize
from bzv.solr import Solr
from bzv.checkpart import CheckPart
from bzv.rss import Rss
from bzv.utils import HandleRequestPrinter, LimitQuery, Printer
from bzv.openurl import OpenURL
from bzv.composedaifrommods import ComposeDaiFromMods
from bzv.composefulljournaltitlefrommods import ComposeFullJournalTitleFromMods
from bzv.composejelfrommods import ComposeJELFromMods
from bzv.composesetsfromheader import ComposeSetsFromHeader
from bzv.composedatestampfromheader import ComposeDateStampFromHeader
from bzv.record import Record, RecordXml, Citec, RecordJSON
from bzv.json import Request, Response
from bzv.repec import RepecRequest
from bzv.googlelikequery import GoogleLikeQuery
from bzv.xmldocumentadapter import XmlDocumentAdapter
from bzv.writedeleted import WriteDeleted
from bzv.undelete import Undelete
from bzv.getavailability import GetAvailability
from bzv.aselect import Aselect
from bzv.storagecomponent import BZVStorageComponent
from bzv.observablewebquery import ObservableWebQuery

def optimize(databasePath):
    weggegooid = False
    indexPath = join(databasePath, 'index')
    for fs in ['write.lock', 'tracker.segments', '*.docids', '*.deleted']:
        for f in iglob(join(indexPath, fs)):
            unlink(f)
            weggegooid = True

    if not weggegooid:
        return

    unlock(indexPath)
    writer = IndexWriter(indexPath, merescoStandardAnalyzer, False)
    writer.optimize()
    writer.close()

class SpyingEye(Observable):
    def __init__(self, infra):
        Observable.__init__(self)
        self.infra = infra
        self.logger = infra.logger
        self.logger.info("Spying eye")
        self.pp = pprint.PrettyPrinter(indent=2)

    def spy(self):
        return self.logger.debug

    def spit(self,text):
        self.logger.debug("Spit: >>>>>>>>>>>>>>>>\n" + self.pp.pformat(text) + "\n<<<<<<<<<<<<<<<<"  )


class GarbageCollector(Observable):
    def handleRequest(self, Client=None, *args, **kwargs):
        gc.collect()
        return self.all.handleRequest(Client=Client, *args, **kwargs)

class ServerLog(Observable, Logger):
    def handleRequest(self, Client=None, *args, **kwargs):
        if Client:
            self.log(ip=Client[0])
        self.log(request='request')
        return self.all.handleRequest(Client=Client, *args, **kwargs)

class OurLogPrinter(Observable):
    def __init__(self, name):
        Observable.__init__(self)
        self._name = name

    def unknown(self, *args, **kwargs):
        print self._name, args, kwargs
        return self.all.unknown(*args, **kwargs)

class StatusReduce(Observable):
    def addField(self, name, value):
        if len(value) == 0:
            value = 'done'
        else:
            value = reduce(lambda a,b: a if a == 'done' else b, value)
        if value == '':
            value = 'failed'
        return self.do.addField(name=name, value=value)

def createUploadHelix(index, storageComponent, xsltFiles, reactor, infra, namespaces):
    yearNormalize = YearNormalize(yearRange=(1000,2100))
    firstInitialNormalize = FirstInitialNormalize()
    lastnameNormalize = LastnameNormalize()
    logger=infra.logger

    serverConfig =  infra.configs['server']
    timeout = int(serverConfig.get('timeout',section='batchTransactions',default="1"))
    batchSize = int(serverConfig.get('batchSize',section='batchTransactions',default="50"))

    DRILLDOWN_PREFIX = serverConfig.get('DRILLDOWN_PREFIX', section = 'drilldown')
    SORT_PREFIX = serverConfig.get('SORT_PREFIX', section = 'drilldown')

    untokenized = []
    untokenizedDict = serverConfig.getSectionDictList('untokenizedFields')

    for key in untokenizedDict:
        untokenized.extend(untokenizedDict[key])

    fields2LuceneDocument = \
        (ResourceManager("record",
                         lambda resourceManager:
                             Fields2LuceneDocumentTx(resourceManager,
                                                     untokenized=untokenized)),
         index
    )

    indexingHelix = \
        (Transparant(),
            fields2LuceneDocument,
            (FilterField(lambda name: DRILLDOWN_PREFIX + name in  untokenizedDict['drilldownfieldnames']),
                (RenameField(lambda name: DRILLDOWN_PREFIX + name),
                    fields2LuceneDocument
                )
            ),
            (FilterField(lambda name: SORT_PREFIX + name in untokenizedDict['sortfieldnames']),
                (RenameField(lambda name: SORT_PREFIX + name),
                    fields2LuceneDocument
                )
            ),
        )

    normalizeTransactionStoreHelix = \
        (ResourceManager("record", lambda resourceManager:Fields2XmlTx(resourceManager, 'extra', namespace="http://meresco.com/namespace/fields/extra")),
            (storageComponent,)
        )

    storeModsHelix = \
        (RewritePartname('mods'),
            (XmlPrintLxml(),
                (storageComponent,)
            ),
        )

    indexModsHelix = \
        (XmlXPath(['//mods:mods'], namespaceMap=namespaces),
            (Xml2Fields(),
                indexingHelix,
                (RenameField(lambda name: name.split('.')[0]),
                    indexingHelix
                ),
                (FilterField(lambda name: name == '__id__'),
                    normalizeTransactionStoreHelix
                )
            ),
            (ComposeNameFromMods('norm.name.oneinitial'),
                (TransformFieldValue(firstInitialNormalize.normalize),
                    indexingHelix,
                    normalizeTransactionStoreHelix,
                ),
                (TransformFieldValue(firstInitialNormalize.unparsable),
                    (RenameField(lambda name: 'norm.name.unparsable'),
                        indexingHelix
                    )
                ),
                (TransformFieldValue(lastnameNormalize.normalize),
                    (RenameField(lambda name: 'norm.name.lastname'),
                        indexingHelix,
                        normalizeTransactionStoreHelix,
                    )
                )
            ),
            (ComposeDaiFromMods('norm.dai'),
                indexingHelix,
                normalizeTransactionStoreHelix,
            ),
            (ComposeFullJournalTitleFromMods('norm.fulljournaltitle'),
                indexingHelix,
                normalizeTransactionStoreHelix,
            ),
            (ComposeJELFromMods('norm.jel'),
                indexingHelix,
                normalizeTransactionStoreHelix,
            ),
#            (XPath2Field([('//mods:location/mods:url/text()', 'fulltext.url.exists')], namespaceMap=namespaces),
#                (TransformFieldValue(lambda value: 'true'),
#                    indexingHelix
#                )
#            ),
            (XPath2Field([('//mods:originInfo/mods:dateIssued/text()', 'norm.date.year')], namespaceMap=namespaces),
                (TransformFieldValue(yearNormalize.normalize),
                    indexingHelix,
                    normalizeTransactionStoreHelix,
                    (TransformFieldValue(lambda year: year[:len('YYY')]+'0'),
                        (RenameField(lambda name: 'norm.date.decennium'),
                            indexingHelix,
                            normalizeTransactionStoreHelix
                        )
                    ),
                    (TransformFieldValue(lambda year: year[:len('YY')]+'00'),
                        (RenameField(lambda name: 'norm.date.centennium'),
                            indexingHelix,
                            normalizeTransactionStoreHelix
                        )
                    )
                ),
                (TransformFieldValue(yearNormalize.unparsable),
                    (RenameField(lambda name: 'norm.date.unparsable'),
                        indexingHelix
                    )
                )
            ),
            (ModsDateField(),
                indexingHelix,
            ),
        )

    return \
        (BatchTransactionScope("batch", reactor=reactor,
                               batchSize=batchSize,
                               timeout=timeout),
            (TransactionScope("record"),
                (Venturi(
                        should=[
                             ('metadata', '/document:document/document:part[@name="metadata"]/text()'),
                             ('meta', '/document:document/document:part[@name="meta"]/text()')],
                        could=[
                            ('rating',   '/rating:rating'),
                            ('header',   '/document:document/document:part[@name="header"]/text()'),
                            ('fulltext', '/fulltext:fulltext'),
                            ('ddi',      '/document:document/document:part[@name="ddi"]/text()'),
                            ('jel',      '/document:document/document:part[@name="jel"]/text()')],
                        namespaceMap = namespaces
                    ),
                    (FilterMessages(allowed=['delete']),
                        index,
                        (WriteDeleted(),
                        (storageComponent,))
                    ),
                    (FilterMessages(allowed=['getStream']),
                        (storageComponent,)
                    ),

################### TP: storing of the required ('should') and optional ('could') parts
                    (XmlPrintLxml(),
                         (storageComponent,)
                    ),

                    (XmlXPath(['//mods:mods'], namespaceMap=namespaces),
                        indexModsHelix
                    ),

                    (XPath2Field([("//oai:metadata/didl:DIDL/didl:Item/didl:Descriptor/didl:Statement/dii:Identifier/text()", 'norm.docid')], namespaceMap=namespaces),
                        indexingHelix,
                        normalizeTransactionStoreHelix,
                    ),
                    ### pica+ ###
                    (XmlXPath(['/oai:metadata//record']),
                        (XsltCrosswalk(xsltFiles),
                            storeModsHelix,
                            indexModsHelix
                         )
                    ),
                    ### fulltext indexing ###
                    (XPath2Field([("//didl:Component/didl:Resource[@mimeType='application/pdf']/@ref", 'fulltext.url.exists')], namespaceMap=namespaces),
                        (TransformFieldValue(lambda value: 'true'),
                            indexingHelix
                        )
                    ),
                    (XmlXPath(['/fulltext:fulltext'], namespaceMap=namespaces),
                        (RewritePartname('fulltext'),
                            (XmlPrintLxml(),
                                 (storageComponent,)
                            ),
                        ),
                        (Xml2Fields(),
                            indexingHelix
                        ),
                        (XPath2Field([('//fulltext:status/text()', 'fulltext.item.status')], namespaceMap=namespaces, sendAsList=True),
                            (StatusReduce(),
                                indexingHelix
                            ),
                        ),
                    ),
                    ## set from header
                    (XmlXPath(['/oai:header'], namespaceMap=namespaces),
                        #(Printer('header-set'),),
                        (ComposeSetsFromHeader('set'),
                            indexingHelix,
                            normalizeTransactionStoreHelix
                        )
                    ),
                    ## datestamp form header
                    (XmlXPath(['/oai:header'], namespaceMap=namespaces),
                        #(Printer('header-datestamp'),),
                        (ComposeDateStampFromHeader('datestamp'),
                            indexingHelix,
                            normalizeTransactionStoreHelix
                        )
                    ),
                    (XmlXPath(['/meta:meta']),
                        (Xml2Fields(),
                            indexingHelix,
                            (FilterField(lambda name: name == 'meta.repository.id'),
                                (RenameField(lambda name: 'meresco.exists'),
                                    (TransformFieldValue(lambda value: 'true'),
                                        indexingHelix
                                    )
                                )
                            ),
                            (RenameField(lambda name: 'meta'),
                                indexingHelix
                            )
                        ),
                    ),
                    (FilterMessages(disallowed=['delete']),
                        (Undelete(),
                        (storageComponent,))
                    )
                )
            )
        )

def main(reactor, observableHttpServer, sruUpdateServer, inbox, baseURL, portNumber, databasePath, configDict, infra, configWatcher):
    logger = infra.logging.getLogger('meresco.server')

    configs =  infra.configs
    websiteConfig = configs ['website']
    serverConfig  = configs ['server']

    appid = configDict['AselectID']

    drilldownFieldnames = serverConfig.getList ('drilldownFieldnames', section='untokenizedFields')
    DRILLDOWN_PREFIX = serverConfig.get('DRILLDOWN_PREFIX', section = 'drilldown')
    namespaces = serverConfig.getSectionDict('namespaces')
    unqualifiedTermFields = serverConfig.getSectionTuple('unqualifiedTermFields')

    applicationName = websiteConfig.get('applicationName',section='userFeedback', default='Beter Zoeken en Vinden')
    storageDirectory = join(databasePath, 'storage')
    indexDirectory = join(databasePath, 'index')

    dynamicPath = configDict['dynamicPath']
    staticPath  = configDict['staticPath']
    xsltPath    = configDict['xsltPath']
    templatePath =configDict['templatePath']
    logPath     = configDict['logPath']

    luceneDebugLogFilename = logPath + '/lucene.log'

    # space seperated filelist
    xsltFileList = configDict['xsltFiles'].split()
    xsltFiles = [join(xsltPath, f) for f in xsltFileList ]

    unlock(indexDirectory)

    storageComponent = BZVStorageComponent(storageDirectory, partsRemovedOnDelete=[])
    drilldownComponent = Drilldown(drilldownFieldnames, transactionName="batch")

    indexHelix = \
            (LuceneIndex(indexDirectory, transactionName="batch",debugLogFilename = luceneDebugLogFilename ),
                (drilldownComponent,)
            )

    uploadHelix = createUploadHelix(indexHelix, storageComponent, xsltFiles, reactor=reactor,infra=infra, namespaces=namespaces)

    configurationComponent = Configuration(configDict)
    rateRecordList = RateRecordCookie(join(databasePath, 'ratedRecords'))

    statisticsPath = join(databasePath, 'statistics')
    isdir(statisticsPath) or makedirs(statisticsPath)
#FIXME statistics in 2.22 aangepast; even uitgezet
#    statistics = Statistics(statisticsPath, [('ip',), ('clause',)], 60)

    locator = serverConfig.get('Locator',section='proxies')
    if serverConfig.get('Testing', section='locator') == 'True':
        print "Testing locator"
        locator = "/availability" + serverConfig.get('Locator',section='proxies')
    else:
        print "Using acceptation/production locator"

    record = \
        (Record(infra),
            (storageComponent,),
            (Record(infra),
                (storageComponent,)
            )
        )

    sru = \
        (SruParser(host=baseURL, port=portNumber, defaultRecordSchema='metadata', defaultRecordPacking='xml'),
            (SruHandler(),
                (CQL2LuceneQuery(unqualifiedTermFields),
                    indexHelix
                ),
                (storageComponent,),
                (SRUTermDrilldown(True),
                    (DrilldownFieldnames(
                        lambda field: DRILLDOWN_PREFIX + field,
                        lambda field: field[len(DRILLDOWN_PREFIX):]),
                            (drilldownComponent,)
                    ),
                    (CQL2LuceneQuery(unqualifiedTermFields),
                        indexHelix
                    ),
                )
            )
        )
    multiLevelDrilldown = \
        (MultiLevelDrilldown(),
            (SpyingEye(infra),),
            (DrilldownConfig(infra),),
            (drilldownComponent,),
         )

    return \
        (Observable(),
            (configWatcher,
            ),
            (inbox,
                (XmlDocumentAdapter(),
                    uploadHelix,
                )
            ),
            (sruUpdateServer,
                (PathFilter('/update'),
                    (SRURecordUpdate(),
                        (Amara2Lxml(),
                            uploadHelix
                        )
                    )
                ),
                (PathFilter('/reindex'),
                    (ReindexConsole(),
                        (Reindex('meta'),
                            (FilterMessages(allowed=['listIdentifiers']),
                                (storageComponent,)
                            ),
                            (FilterMessages(disallowed=['listIdentifiers']),
                                uploadHelix
                             )
                        )
                    )
                )
            ),
            (observableHttpServer,
# FIXME statistics is veranderd in 2.22; even uitgezet
#                (statistics,
#               (Aselect('^/$', baseURL, 'foobar'),
                    (SpyingEye(infra),),
                    (ServerLog(), ),
#                    (PathFilter('/statistics'),
#                        (StatisticsXml(statistics), )
#                    ),
                    (PathFilter('/rss'),
                        (Rss(   title = applicationName,
                                description = 'Tilburg University',
                                link = baseURL,
                                maximumRecords = 15),
                            (CQL2LuceneQuery(unqualifiedTermFields),
                                indexHelix
                            ),
                            (CheckPart(parts=['mods'], action='stop'),
                                (storageComponent,),
                                (RssItem(
                                    nsMap=namespaces,
                                    title = ('metadata', '//mods:mods/mods:titleInfo/mods:title/text()'),
                                    description = ('metadata', '//mods:mods/mods:abstract/text()'),
                                    linkTemplate = baseURL + '/page/recordview?recordId=%(uploadId)s',
                                    uploadId = ('meta', '/meta:meta/meta:upload/meta:id/text()')),
                                    (storageComponent,)
                                ),
                                (CheckPart(parts=['mods'], action='continue'),
                                    (storageComponent,),
                                    (RssItem(
                                        nsMap=namespaces,
                                        title = ('mods', '//mods:mods/mods:titleInfo/mods:title/text()'),
                                        description = ('mods', '//mods:mods/mods:abstract/text()'),
                                        linkTemplate = baseURL + '/page/recordview?recordId=%(uploadId)s',
                                        uploadId = ('meta', '/meta:meta/meta:upload/meta:id/text()')),
                                       (storageComponent, )
                                    )
                                )
                            )
                        )
                    ),
                    (PathFilter('/record'),
                        (RecordXml(),
                            record
                        )
                    ),
                    (PathFilter('/citec'),
                        (Citec(),
                            record
                        )
                    ),
                    (PathFilter('/JSON/record'),
                        (RecordJSON(),
                            record
                        )
                    ),
                    (PathFilter('/JSON/searchresult'),
                        (HandleRequestPrinter(),),
                        (Request(configs),
                            (GoogleLikeQuery(), ),
                            (CQL2LuceneQuery(unqualifiedTermFields),
                                indexHelix
                            ),
                            (Record(infra),
                                (storageComponent,)
                            ),
                            (drilldownComponent,),
                            multiLevelDrilldown,

                            (SpyingEye(infra),),
                            (DrilldownConfig(infra),),

                        )
                    ),
                    (PathFilter('/repec'),
                        (HandleRequestPrinter(),),
                        (RepecRequest(configs, xsltFiles),
                            (CQL2LuceneQuery(unqualifiedTermFields),
                                indexHelix
                            ),
                            (storageComponent,)
                        )
                    ),
################### TP: this is only for testing
                    (PathFilter('/availability'),
                        (GetAvailability(configs),)
                    ),
                    (PathFilter('/eo/sru'),
                        (HandleRequestPrinter(),),
                        (LimitQuery(" AND (meta.repository.collection exact neeo OR meta.repository.collection exact repec)"),
                            sru
                        )
                    ),
                    (PathFilter('/sru'),
                        (HandleRequestPrinter(),),
                        sru
                    ),
                    (PathFilter('/files'),\
                        (PathRename(lambda path: path[len('/files'):]),
                            (FileServer(staticPath),)
                        )
                    ),
                    (PathFilter('/gc'),\
                        (GarbageCollector(),)
                    ),

                    (PathFilter(['/', '/m/search', '/search', '/id'] , excluding=[ '/rss', '/record', '/citec', '/JSON/record', '/JSON/searchresult', '/repec', '/availability', '/eo/sru', '/sru', '/update', '/files', '/statistics', '/reindex', '/gc']),
                        (DynamicHtml([dynamicPath], reactor=reactor, allowedModules=['cqlparser', 'cqlwebutils'],
                                     indexPage='/search',
                                     infra = infra,
                                     configs = configs,
                                     templatePath = templatePath,
                                     ),
                            (SpyingEye(infra),),
                            (DrilldownConfig(infra),),

                            (GoogleLikeQuery(), ),
                            (CQL2LuceneQuery(unqualifiedTermFields),
#                                (Solr("http://acceptatie.solr.uvt.nl/solr/select/"),),
                                indexHelix
                            ),
                            (Response(), ),
                            (OpenURL(locator),),
                            record,
                            (storageComponent,),
                            (drilldownComponent,),
                            (RateRecord(1,5),
                                (FilterMessages(allowed=['isAvailable', 'getStream']),
                                    (storageComponent,),
                                ),
                                uploadHelix
                            ),
                            multiLevelDrilldown,
                            (configurationComponent,),
                            (rateRecordList, ),
                        )
                    )


#                )
            )
        )
